

/***********************************************************************

   This software is for research and educational purposes only.

************************************************************************/



#ifndef ntl_vector__H
#define ntl_vector__H

#include "tools.h"

struct _VectorHeader {
   long length;
   long alloc;
   long init;
   long fixed;
};

union _Aligned_VectorHeader {
   _VectorHeader h;
   double x1;
   long x2;
   char *x3;
};

#define NTL_VEC_HEAD(p) (& (((_Aligned_VectorHeader *) p)[-1].h))


inline void *operator new(size_t, void *p) { return p; }

#ifndef NTL_RANGE_CHECK
#define NTL_RANGLE_CHECK_CODE 
#else
#define NTL_RANGLE_CHECK_CODE if (i < 0 || !rep || i >= NTL_VEC_HEAD(rep)->length) RangeError(i);
#endif

// vectors are allocated in chunks of this size

#ifndef NTL_VectorMinAlloc
#define NTL_VectorMinAlloc (4)
#endif

// vectors are always expanded by at least this ratio

#ifndef NTL_VectorExpansionRatio
#define NTL_VectorExpansionRatio (1.2)
#endif

// controls initialization during input

#ifndef NTL_VectorInputBlock
#define NTL_VectorInputBlock 50
#endif


#define ntl_vector_default(T)   \
void BlockConstruct(T* p, long n)  \
{  \
   long i;  \
  \
   for (i = 0; i < n; i++)  \
      (void) new(&p[i]) T;  \
}  \
  \
void BlockDestroy(T* p, long n)  \
{  \
   long i;  \
  \
   for (i = 0; i < n; i++)  \
      p[i].~T();  \
}



#define ntl_vector_decl(T,vec_T)  \
class vec_T {  \
private:  \
   T *rep;  \
  \
   void RangeError(long i) const;  \
  \
public:  \
   vec_T() { rep = 0; }  \
   vec_T(INIT_SIZE_TYPE, long n) { rep = 0; SetLength(n); }  \
   vec_T(const vec_T& a) { rep = 0; *this = a; }     \
   vec_T& operator=(const vec_T& a);  \
   ~vec_T();  \
   void kill(); \
  \
   void SetLength(long n);  \
   void SetMaxLength(long n); \
   void FixLength(long n); \
   void QuickSetLength(long n) { NTL_VEC_HEAD(rep)->length = n; } \
  \
   long length() const { return (!rep) ?  0 : NTL_VEC_HEAD(rep)->length; }  \
   long MaxLength() const { return (!rep) ?  0 : NTL_VEC_HEAD(rep)->init; } \
   long fixed() const { return rep && NTL_VEC_HEAD(rep)->fixed; } \
  \
   T& operator[](long i)   \
   {  \
      NTL_RANGLE_CHECK_CODE  \
      return rep[i];  \
   }  \
  \
   const T& operator[](long i) const \
   {  \
      NTL_RANGLE_CHECK_CODE  \
      return rep[i];  \
   }  \
  \
   T& RawGet(long i)   \
   {  \
      return rep[i];  \
   }  \
  \
   const T& RawGet(long i) const \
   {  \
      return rep[i];  \
   }  \
  \
   T& operator()(long i) { return (*this)[i-1]; }  \
   const T& operator()(long i) const { return (*this)[i-1]; } \
   \
   friend void swap(vec_T& x, vec_T& y);  \
  \
   const T* elts() const { return rep; }  \
   T* elts() { return rep; }  \
         \
   friend void append(vec_T& v, const T& a); \
   friend void append(vec_T& v, const vec_T& w); \
 \
   vec_T(vec_T& x, INIT_TRANS_TYPE) { rep = x.rep; x.rep = 0; } \
   long position(const T& a) const;  \
}; 




#define ntl_io_vector_decl(T,vec_T)  \
istream& operator>>(istream&, vec_T&);  \
  \
ostream& operator<<(ostream&, const vec_T&);  \


#define ntl_eq_vector_decl(T,vec_T)  \
long operator==(const vec_T& a, const vec_T& b);  \
long operator!=(const vec_T& a, const vec_T& b);


#define ntl_vector_impl(T,vec_T) ntl_vector_default(T) ntl_vector_impl_plain(T,vec_T)  

#define ntl_vector_impl_plain(T,vec_T)  \
 \
void vec_T::SetLength(long n)   \
{   \
   long m;  \
  \
   if (n < 0) {  \
      Error("negative length in vector::SetLength");  \
   }  \
      \
   if (rep && NTL_VEC_HEAD(rep)->fixed) {\
      if (NTL_VEC_HEAD(rep)->length == n) \
         return; \
      else \
         Error("SetLength: can't change this vector's length"); \
   }  \
   if (n == 0) {  \
      if (rep) NTL_VEC_HEAD(rep)->length = 0;  \
      return;  \
   }  \
  \
   if (!rep) {  \
      m = ((n+NTL_VectorMinAlloc-1)/NTL_VectorMinAlloc) * NTL_VectorMinAlloc; \
      char *p = (char *) malloc(sizeof(_Aligned_VectorHeader)+sizeof(T)*m); \
      if (!p) {  \
	 Error("out of memory in vector::SetLength()");  \
      }  \
      rep = (T *) (p + sizeof(_Aligned_VectorHeader)); \
  \
      BlockConstruct(rep, n); \
  \
      NTL_VEC_HEAD(rep)->length = n;  \
      NTL_VEC_HEAD(rep)->init = n;  \
      NTL_VEC_HEAD(rep)->alloc = m;  \
      NTL_VEC_HEAD(rep)->fixed = 0;  \
   }  \
   else if (n <= NTL_VEC_HEAD(rep)->init) {  \
      NTL_VEC_HEAD(rep)->length = n;  \
   }  \
   else  {  \
      if (n > NTL_VEC_HEAD(rep)->alloc) {  \
	 if (n < NTL_VectorExpansionRatio*NTL_VEC_HEAD(rep)->alloc)  \
	    m = (long) (NTL_VectorExpansionRatio*NTL_VEC_HEAD(rep)->alloc);  \
         else  \
	    m = n;  \
         m = ((m+NTL_VectorMinAlloc-1)/NTL_VectorMinAlloc) * NTL_VectorMinAlloc; \
         char *p = ((char *) rep) - sizeof(_Aligned_VectorHeader); \
         p = (char *) realloc(p, sizeof(_Aligned_VectorHeader)+sizeof(T)*m); \
         if (!p) {  \
	    Error("out of memory in vector::SetLength()");  \
         }  \
         rep = (T *) (p + sizeof(_Aligned_VectorHeader)); \
	 NTL_VEC_HEAD(rep)->alloc = m;  \
      }  \
      BlockConstruct(rep + NTL_VEC_HEAD(rep)->init, n - NTL_VEC_HEAD(rep)->init); \
      NTL_VEC_HEAD(rep)->length = n;  \
      NTL_VEC_HEAD(rep)->init = n;  \
   }  \
}  \
 \
 \
void vec_T::SetMaxLength(long n) \
{ \
   long OldLength = length(); \
   SetLength(n); \
   SetLength(OldLength); \
} \
 \
void vec_T::FixLength(long n) \
{ \
   if (rep) Error("FixLength: can't fix this vector"); \
   if (n < 0) Error("FixLength: negative length"); \
   if (n > 0) \
      SetLength(n); \
   else { \
      char *p = (char *) malloc(sizeof(_Aligned_VectorHeader)); \
      if (!p) {  \
	 Error("out of memory in vector::FixLength()");  \
      }  \
      rep = (T *) (p + sizeof(_Aligned_VectorHeader)); \
  \
      NTL_VEC_HEAD(rep)->length = 0;  \
      NTL_VEC_HEAD(rep)->init = 0;  \
      NTL_VEC_HEAD(rep)->alloc = 0;  \
   } \
   NTL_VEC_HEAD(rep)->fixed = 1; \
} \
  \
vec_T& vec_T::operator=(const vec_T& a)  \
{  \
   long i, n;  \
   T *p;  \
   const T *ap;  \
  \
   n = a.length();  \
   SetLength(n);  \
   ap = a.elts();  \
   p = elts();  \
  \
   for (i = 0; i < n; i++)  \
      p[i] = ap[i];  \
   return *this;  \
}  \
       \
  \
vec_T::~vec_T()  \
{  \
   if (!rep) return;  \
   BlockDestroy(rep, NTL_VEC_HEAD(rep)->init); \
   free(((char *) rep) - sizeof(_Aligned_VectorHeader));  \
}  \
   \
void vec_T::kill()  \
{  \
   if (!rep) return;  \
   if (NTL_VEC_HEAD(rep)->fixed) Error("can't kill this vector"); \
   BlockDestroy(rep, NTL_VEC_HEAD(rep)->init); \
   free(((char *) rep) - sizeof(_Aligned_VectorHeader));  \
   rep = 0; \
}  \
  \
void vec_T::RangeError(long i) const  \
{  \
   cerr << "index out of range in vector: ";  \
   cerr << i;  \
   if (!rep)  \
      cerr << "(0)\n";  \
   else  \
      cerr << "(" << NTL_VEC_HEAD(rep)->length << ")\n";  \
   abort();  \
}  \
  \
long vec_T::position(const T& a) const  \
{  \
   if (!rep) return -1;  \
   long num_alloc = NTL_VEC_HEAD(rep)->alloc;  \
   long num_init = NTL_VEC_HEAD(rep)->init;  \
   if (&a < rep || &a >= rep + num_alloc) return -1;  \
   long res = (&a) - rep;  \
   if (res > num_init) Error("position: reference to uninitialized object"); \
   return res;  \
}  \
 \
void swap(vec_T& x, vec_T& y)  \
{  \
   T* t;  \
   long xf = x.fixed();  \
   long yf = y.fixed();  \
   if (xf != yf ||   \
       (xf && NTL_VEC_HEAD(x.rep)->length != NTL_VEC_HEAD(y.rep)->length))  \
      Error("swap: can't swap these vectors");  \
   t = x.rep;  \
   x.rep = y.rep;  \
   y.rep = t;  \
} \
 \
void append(vec_T& v, const T& a)  \
{  \
   long l = v.length(); \
   long pos = v.position(a);  \
   v.SetLength(l+1);  \
   if (pos != -1)  \
      v[l] = v.RawGet(pos);  \
   else  \
      v[l] = a;  \
}  \
  \
void append(vec_T& v, const vec_T& w)  \
{  \
   long l = v.length();  \
   long m = w.length();  \
   long i;  \
   v.SetLength(l+m);  \
   for (i = 0; i < m; i++)  \
      v[l+i] = w[i];  \
}





#define ntl_io_vector_impl(T,vec_T)  \
istream & operator>>(istream& s, vec_T& a)   \
{   \
   vec_T ibuf;  \
   long c;   \
   long n;   \
   if (!s) Error("bad vector input"); \
   \
   c = s.peek();  \
   while (c == ' ' || c == '\n' || c == '\t') {  \
      s.get();  \
      c = s.peek();  \
   }  \
   if (c != '[') {  \
      Error("bad vector input");  \
   }  \
   \
   n = 0;   \
   ibuf.SetLength(0);  \
      \
   s.get();  \
   c = s.peek();  \
   while (c == ' ' || c == '\n' || c == '\t') {  \
      s.get();  \
      c = s.peek();  \
   }  \
   while (c != ']' && c != EOF) {   \
      if (n % NTL_VectorInputBlock == 0) ibuf.SetMaxLength(n + NTL_VectorInputBlock); \
      n++;   \
      ibuf.SetLength(n);   \
      if (!(s >> ibuf[n-1])) Error("bad vector input");   \
      c = s.peek();  \
      while (c == ' ' || c == '\n' || c == '\t') {  \
         s.get();  \
         c = s.peek();  \
      }  \
   }   \
   if (c == EOF) Error("bad vector input");  \
   s.get(); \
   \
   a = ibuf; \
   return s;   \
}    \
   \
   \
ostream& operator<<(ostream& s, const vec_T& a)   \
{   \
   long i, n;   \
  \
   n = a.length();  \
   \
   s << '[';   \
   \
   for (i = 0; i < n; i++) {   \
      s << a[i];   \
      if (i < n-1) s << " ";   \
   }   \
   \
   s << ']';   \
      \
   return s;   \
}   \

#define ntl_eq_vector_impl(T,vec_T) \
long operator==(const vec_T& a, const vec_T& b) \
{  \
   long n = a.length();  \
   if (b.length() != n) return 0;  \
   const T* ap = a.elts(); \
   const T* bp = b.elts(); \
   long i;  \
   for (i = 0; i < n; i++) if (ap[i] != bp[i]) return 0;  \
   return 1;  \
} \
long operator!=(const vec_T& a, const vec_T& b) \
{  return !(a == b); }

   


#endif

